"""User model"""
import uuid
import hashlib

from time import time

from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import relationship, joinedload
from sqlalchemy.orm.session import Session
from sqlalchemy.types import String, Unicode, Integer, Boolean

from . import Base, DBSession


def md5(s):
    return hashlib.md5(s.encode('utf8')).hexdigest()


def sha1(s):
    return hashlib.sha1(s.encode('utf8')).hexdigest()


class User(Base):
    __tablename__ = 'pbuser'
    __table_args__ = {'mysql_charset': 'utf8', 'mysql_engine': 'InnoDB'}

    _str_roles = None

    id = Column(Integer, primary_key=True)
    login = Column(String(255))
    password = Column(String(255))
    display_name = Column(Unicode(255))
    email = Column(Unicode(255))
    # user kind (class, type), possible values: "local", "twitter"
    kind = Column(String(20))

    roles = relationship('Role')

    def detach(self):
        dbsession = Session.object_session(self)
        if dbsession is None:
            return

        for x in self.roles:
            dbsession.expunge(x)

        dbsession.expunge(self)

    def has_role(self, r):
        return r in self.get_roles()

    def get_roles(self):
        if self._str_roles is None:
            self._str_roles = [x.name for x in self.roles]

        return self._str_roles


class AnonymousUser:
    kind = 'anonymous'

    def has_role(self, r):
        return False

anonymous = AnonymousUser()


class Role(Base):
    __tablename__ = 'pbuserrole'
    __table_args__ = {'mysql_charset': 'utf8', 'mysql_engine': 'InnoDB'}

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('pbuser.id'))
    name = Column(String(50))

    def __init__(self, id, user_id, name):
        self.id = id
        self.user_id = user_id
        self.name = name


class VerifiedEmail(Base):
    __tablename__ = 'pbverifiedemail'
    __table_args__ = {'mysql_charset': 'utf8', 'mysql_engine': 'InnoDB'}

    # stripperd lowcased email address
    id = Column(Integer, primary_key=True)
    email = Column(Unicode(255), unique=True)
    is_verified = Column(Boolean)
    last_verify_date = Column(Integer)
    verification_code = Column(String(255))

    def __init__(self, email):
        self.last_verify_date = int(time())
        self.email = email
        self.is_verified = False
        self.verification_code = str(uuid.uuid4())


def find_local_user(login, password):
    hashed_password = md5(password)
    dbsession = DBSession()
    q = dbsession.query(User).options(joinedload('roles')).filter(User.kind == 'local').\
        filter(User.login == login)
    user = q.first()

    if user.password == hashed_password:
        # old password hashing method is used
        return user
    else:
        # possibly a new method is used
        salt = user.password[:8]
        hashed_password = salt + sha1(salt + sha1(password))
        if user.password == hashed_password:
            return user

    return None


def normalize_email(email):
    return email


def get_user(user_id):
    dbsession = DBSession()
    user = dbsession.query(User).options(joinedload('roles')).get(user_id)
    return user


def find_twitter_user(username):
    dbsession = DBSession()
    q = dbsession.query(User).options(joinedload('roles')).filter(User.kind == 'twitter').\
        filter(User.login == username)
    user = q.first()
    return user



"""
Notifications
"""

import logging
import urllib
import re
from smtplib import SMTP
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from multiprocessing import Process


from pyramid.url import route_url
from pyramid import threadlocal
from pyramid.renderers import render

from pyrone.models.config import get as get_config
from pyrone.models.user import normalize_email
from pyrone.lib.lang import lang
from pyrone.lib import helpers as h

log = logging.getLogger(__name__)

enable_email = False
smtp_server = None


def init_notifications_from_settings(settings):
    global enable_email, smtp_server

    if settings.get('pyrone.notifications.mail') != 'true':
        return

    m_transport = settings.get('pyrone.notifications.mail_transport')
    if m_transport not in ('smtp',):
        log.error('Not supported email transport "{0}"'.format(m_transport))
        return

    enable_email = True
    smtp_server = settings.get('pyrone.notifications.mail_smtp_server')

COMMASPACE = ', '
SUBJECT_RE = re.compile(r'^Subject: ([^\n\r]+)')

def send_email_process(request, template_name, recipients, sender, params):
    """Process for real sending emails"""
    language = lang(request)

    html = render('/email/'+template_name+'.mako', params, request)

    # now cut out subject line from the message
    mo = SUBJECT_RE.match(html)
    if mo is None:
        subject = 'NO-SUBJECT'
    else:
        subject = mo.group(1)
        html = SUBJECT_RE.sub('', html, 1)

    msg = MIMEMultipart()
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = COMMASPACE.join(recipients)
    msg.preamble = 'Use multipart, Luke'

    html_part = MIMEText(html, 'html', 'utf-8')
    msg.attach(html_part)

    log.debug('--------------------------------------------------')
    debug_data = '''
        SEND NOTIFICATION:
        Subject: {subject}
        To: {to}

        {body}
        '''.format(subject=subject, to=msg['To'], body=html)
    log.debug(debug_data)
    log.debug('--------------------------------------------------')

    with SMTP(smtp_server) as smtp:
        smtp.send_message(msg)


class Notification:
    """
    Object represents notification object
    """
    to = None
    request = None
    params = None

    def send(self):

        if not enable_email:
            log.debug('mail sending is not allowed in config')
            return

        sender = get_config('notifications_from_email')
        Process(target=send_email_process, args=(self.request, self.template_name, 
            [self.to], sender, self.params)).start()

    def __init__(self, template_name, to, params, request=None):
        self.to = to
        self.template_name = template_name
        self.params = params

        if request is None:
            request = threadlocal.get_current_request()

        self.request = request


def _extract_comment_sub(request, comment):
    """
    Extract placeholders substitution strings from the comment object
    """
    author_name = comment.display_name
    author_email = comment.email

    if comment.user is not None:
        author_name = comment.user.display_name
        author_email = comment.user.email

    # construct comment link
    article = comment.article
    comment_url = h.article_url(request, article) + '#comment-' + str(comment.id)

    comment_date = h.timestamp_to_str(comment.published)
    res = {
        'comment_author_email': author_email,
        'comment_author_name': author_name,
        'comment_text': comment.body,
        'comment_date': comment_date,
        'comment_url': comment_url
        }
    return res


def _extract_article_sub(request, article):
    """
    Extract placeholders substitution strings from the article object
    """
    article_url = h.article_url(request, article)

    res = {
        'article_title': article.title,
        'article_url': article_url
        }
    return res


def gen_comment_response_notification(request, article, comment, top_comment, email):
    """
    Generate comment answer notification
    """

    email = normalize_email(email)
    if not email:
        return None

    # placeholders replacements
    params = {
        'site_title': get_config('site_title')
    }
    params.update(_extract_comment_sub(request, comment))
    params.update(_extract_article_sub(request, article))

    n = Notification('comment_response', email, params, request=request)

    return n


def gen_email_verification_notification(request, email, verification_code):
    """
    Generate email address verification notification.
    """

    email = normalize_email(email)
    if not email:
        return None

    # placeholders replacements
    base_url = get_config('site_base_url')
    q = urllib.parse.urlencode({'token': verification_code, 'email': email})
    verify_url = base_url + '/verify-email?' + q
    params = {
        'site_title': get_config('site_title'),
        'email': email,
        'verify_url': verify_url
    }

    n = Notification('email_verification', email, params, request=request)

    return n


def gen_new_comment_admin_notification(request, article, comment):
    """
    Generate new comment notification for the administrator.
    """
    email = normalize_email(get_config('admin_notifications_email'))
    if not email:
        return None

    # placeholders replacements
    params = {
        'site_title': get_config('site_title')
    }
    params.update(_extract_comment_sub(request, comment))
    params.update(_extract_article_sub(request, comment.article))

    n = Notification('admin_new_comment', email, params, request=request)

    return n



from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPFound
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid_beaker import session_factory_from_settings
from sqlalchemy import engine_from_config

from pyrone.models import initialize_sql
from pyrone.lib.auth import PyroneSessionAuthenticationPolicy, get_user
from pyrone.models.file import init_storage_from_settings
from pyrone.lib.notifications import init_notifications_from_settings
from pyrone.lib.lang import locale_negotiator
#from pyrone.resources import RootFactory

def add_route(config, **r):
    """Adds route with trailing slashes if requested
    """
    is_slashed = '_slashed' in r

    rname = r['name']

    if is_slashed:
        del r['_slashed']
        r['name'] = rname + '_slashed-auto'
        pattern = r['pattern']
        r['pattern'] = pattern + '/'
        config.add_route(**r)

        config.add_view(lambda request: HTTPFound(request.route_url(rname, _query=request.GET, **request.matchdict)),
            route_name=rname + '_slashed-auto')
        r['pattern'] = pattern

    r['name'] = rname
    config.add_route(**r)


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    engine = engine_from_config(settings, 'sqlalchemy.')
    initialize_sql(engine)
    session_factory = session_factory_from_settings(settings)
    init_storage_from_settings(settings)
    init_notifications_from_settings(settings)
    authentication_policy = PyroneSessionAuthenticationPolicy()
    authorization_policy = ACLAuthorizationPolicy()
    config = Configurator(settings=settings, session_factory=session_factory,
        root_factory='pyrone.resources.RootFactory',
        authentication_policy=authentication_policy,
        authorization_policy=authorization_policy,
        locale_negotiator=locale_negotiator)
    config.add_translation_dirs('pyrone:locale/')
    config.set_request_property(get_user, 'user', reify=True)  # reify=True means that result is cached per request
    config.scan()
    config.add_static_view('static', 'pyrone:static')
    routes = [
        {'name': 'static_favicon_png', 'pattern': '/favicon.png'},
        {'name': 'static_favicon_ico', 'pattern': '/favicon.ico'},
        {'name': 'blog_latest', 'pattern': '/'},
        {'name': 'blog_latest_rss', 'pattern': '/rss/latest'},
        {'name': 'blog_tag_articles', 'pattern': '/tag/{tag}', '_slashed': True},
        {'name': 'blog_write_article', 'pattern': '/write'},
        {'name': 'blog_go_article', 'pattern': '/article/{article_id:\d+}'},
        {'name': 'blog_view_article', 'pattern': '/{shortcut_date:\d\d\d\d/\d\d/\d\d}/{shortcut}', '_slashed': True},
        {'name': 'blog_view_moderation_queue', 'pattern': '/comments/moderation'},
        {'name': 'blog_edit_article_ajax', 'pattern': '/article/{article_id:\d+}/edit/ajax'},
        {'name': 'blog_edit_article', 'pattern': '/article/{article_id:\d+}/edit'},
        {'name': 'blog_preview_article', 'pattern': '/preview/article'},
        {'name': 'blog_article_delete_ajax', 'pattern': '/article/{article_id:\d+}/delete/ajax'},
        {'name': 'blog_edit_comment_ajax', 'pattern': '/comment/{comment_id:\d+}/edit/ajax'},
        {'name': 'blog_approve_comment_ajax', 'pattern': '/comment/{comment_id:\d+}/approve/ajax'},
        {'name': 'blog_delete_comment_ajax', 'pattern': '/comment/{comment_id:\d+}/delete/ajax'},
        {'name': 'blog_edit_comment_fetch_ajax', 'pattern': '/comment/{comment_id:\d+}/fetch/ajax'},
        {'name': 'blog_add_article_comment', 'pattern': '/article/{article_id:\d+}/comment'},
        {'name': 'blog_add_article_comment_ajax', 'pattern': '/article/{article_id:\d+}/comment/ajax'},

        {'name': 'account_my_profile', 'pattern': '/me'},
        {'name': 'account_save_my_profile_ajax', 'pattern': '/me/save/ajax'},
        {'name': 'account_login', 'pattern': '/login'},
        {'name': 'account_logout', 'pattern': '/logout'},
        {'name': 'account_verify_email', 'pattern': '/verify-email'},
        {'name': 'account_twitter_init', 'pattern': '/login/twitter/init'},
        {'name': 'account_twitter_finish', 'pattern': '/login/twitter/finish'},
        {'name': 'account_my_subscriptions', 'pattern': '/me/subscriptions'},

        {'name': 'blog_download_file', 'pattern': '/files/f/{filename}'},
        {'name': 'blog_download_file_preview', 'pattern': '/files/p/{filename}'},

        {'name': 'admin_settings', 'pattern': '/admin/settings'},
        {'name': 'admin_settings_save_ajax', 'pattern': '/admin/settings/save/ajax'},
        {'name': 'admin_settings_widget_pages', 'pattern': '/admin/settings/widget/pages'},
        {'name': 'admin_settings_widget_pages_save_ajax', 'pattern': '/admin/settings/widget/pages/save/ajax'},
        {'name': 'admin_list_files', 'pattern': '/admin/files'},
        {'name': 'admin_upload_file', 'pattern': '/admin/file/upload'},
        {'name': 'admin_upload_file_check_ajax', 'pattern': '/admin/file/upload/check'},
        {'name': 'admin_edit_file_props', 'pattern': '/admin/file/{file_id:\d+}/edit'},
        {'name': 'admin_edit_file_props_check_ajax', 'pattern': '/admin/file/{file_id:\d+}/edit/check'},
        {'name': 'admin_delete_files_ajax', 'pattern': '/admin/files/delete/ajax'},
        {'name': 'admin_list_accounts', 'pattern': '/admin/accounts'},
        {'name': 'admin_list_visitors_emails', 'pattern': '/admin/visitors-emails'},
        {'name': 'admin_visitor_email_edit_ajax', 'pattern': '/admin/visitor-email/edit/ajax'},
        {'name': 'admin_visitors_emails_delete_ajax', 'pattern': '/admin/visitors-emails/delete/ajax'},
        {'name': 'admin_delete_accounts_ajax', 'pattern': '/admin/accounts/delete/ajax'},
        {'name': 'admin_list_backups', 'pattern': '/admin/backups'},
        {'name': 'admin_upload_backup', 'pattern': '/admin/backups/upload'},
        {'name': 'admin_delete_backups_ajax', 'pattern': '/admin/backups/delete/ajax'},
        {'name': 'admin_restore_backup', 'pattern': '/admin/backup/{backup_id}/restore'},
        {'name': 'admin_download_backup', 'pattern': '/admin/backup/{backup_id}/download'},
        {'name': 'admin_backup_now', 'pattern': '/admin/backup/now'}
    ]
    for r in routes:
        add_route(config, **r)

    config.add_subscriber('pyrone.subscribers.add_renderer_globals', 'pyramid.events.BeforeRender')
    config.add_subscriber('pyrone.subscribers.add_localizer', 'pyramid.events.NewRequest')

    return config.make_wsgi_app()
